# -*- coding: utf-8 -*-


def filter_duplications(blocks_by_ids):
    for block_id in list(blocks_by_ids.keys()):
        blocks = blocks_by_ids[block_id]
        genomes = [block.parent_seq.genome_name for block in blocks]
        genomes_set = set(genomes)
        if len(genomes_set) != len(genomes):
            del blocks_by_ids[block_id]


def get_all_genomes_from_blocks(blocks_as_ids):
    genomes = set()
    for block_id, blocks in blocks_as_ids.items():
        for block in blocks:
            genomes.add(block.parent_seq.genome_name)
    return genomes


def filter_indels(blocks_by_ids, all_genomes_as_set=None):
    if all_genomes_as_set is None:
        all_genomes_as_set = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    for block_id in list(blocks_by_ids.keys()):
        blocks = blocks_by_ids[block_id]
        genomes_set = {block.parent_seq.genome_name for block in blocks}
        if len(all_genomes_as_set) != len(genomes_set):
            del blocks_by_ids[block_id]


def filter_blocks_by_good_genomes(blocks_by_ids, good_genomes):
    for block_id in list(blocks_by_ids.keys()):
        blocks = blocks_by_ids[block_id]
        new_blocks = [block for block in blocks if block.parent_seq.genome_name in good_genomes]
        if len(new_blocks) == 0:
            del blocks_by_ids[block_id]
        else:
            blocks_by_ids[block_id] = new_blocks


def filter_blocks_by_bad_genomes(blocks_by_ids, bad_genomes):
    for block_id in list(blocks_by_ids.keys()):
        blocks = blocks_by_ids[block_id]
        new_blocks = [block for block in blocks if block.parent_seq.genome_name not in bad_genomes]
        if len(new_blocks) == 0:
            del blocks_by_ids[block_id]
        else:
            blocks_by_ids[block_id] = new_blocks



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import logging
import os
import sys

import configargparse

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
from camsa.core.data_structures import Sequence
from camsa.core.io import write_seqi
from camsa.utils.ragout.shared import get_all_genomes_from_blocks, filter_blocks_by_good_genomes, filter_blocks_by_bad_genomes, filter_indels, filter_duplications
import camsa.utils.ragout.io as ragout_io

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting Ragout coords formatted blocks for further CAMSA processing.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "ragout", "ragout_coords2camsa_seqi.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("-o", "--output-file", metavar="OUTPUT", dest="output", type=configargparse.FileType("wt"), default=sys.stdout,
                        help="A file to which the CAMSA readable fragments lengths would be written. Standard extension is \".camsa.lengths\".\nDEFAULT: stdout")
    parser.add_argument("--o-format", type=str, help="")
    parser.add_argument("ragout_coords", type=str, help="A path to ragout coords file")
    parser.add_argument("--genomes-order", type=str, default="",
                        help="A coma separated list of genome names, used to determine the output order for each homologous synteny block.\nDEFAULT: \"\" (i.e., sorted list of good (all - bad) genomes)")
    parser.add_argument("--ann-genomes", type=str, default="",
                        help="A coma separated list of genome names, to be used when annotation all synteny blocks in the output.\nDEFAULT: \"\" (i.e., equals to the --genomes-order value)")
    parser.add_argument("--ann-delimiter", type=str, default=";")
    parser.add_argument("--filter-indels", action="store_true", dest="filter_indels", default=False,
                        help="A flag to indicate whether to filter out all synteny blocks, that are not present exactly once in each of the good (all-bad) genomes.")
    parser.add_argument("--no-filter-indels", action="store_false", dest="filter_indels", default=False,
                        help="A flag to indicate whether to filter out all synteny blocks, that are not present exactly once in each of the good (all-bad) genomes.")
    parser.add_argument("--filter-duplications", action="store_true", dest="filter_duplications", default=True,
                        help="A flag to indicate whether to filter out all synteny blocks, that are present more than once in at least on of the good (all-bad) genomes.")
    parser.add_argument("--no-filter-duplications", action="store_false", dest="filter_duplications", default=True,
                        help="A flag to indicate whether to filter out all synteny blocks, that are present more than once in at least on of the good (all-bad) genomes.")
    parser.add_argument("--good-genomes", type=str, default="", help="A coma separated list of genome names, to be processed and conversed.\nDEFAULT: \"\" (i.e., all genomes are good)")
    parser.add_argument("--bad-genomes", type=str, default="", help="A coma separated list of genome names, to be excluded from processing and conversion.\nDEFAULT: \"\" (i.e., no genomes are bad)")
    parser.add_argument("--c-logging-level", dest="logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    parser.add_argument("--o-delimiter", type=str, default="\t",
                        help="A single character string, used as a delimiter in the output (t)/(c)sv file.\nDEFAULT: \\t")
    args = parser.parse_args()
    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.ragout_coords2camsa_seqi")
    ch = logging.StreamHandler()
    ch.setLevel(args.logging_level)
    logger.setLevel(args.logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    sequences_by_ids, blocks_by_ids = ragout_io.read_from_file(path=args.ragout_coords, silent_fail=False, delimiter="\t")
    all_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.good_genomes != "":
        args.good_genomes = set(args.good_genomes.split(","))
        filter_blocks_by_good_genomes(blocks_by_ids=blocks_by_ids, good_genomes=args.good_genomes)
    if args.bad_genomes != "":
        args.bad_genomes = set(args.bad_genomes.split(","))
        filter_blocks_by_bad_genomes(blocks_by_ids=blocks_by_ids, bad_genomes=args.bad_genomes)
    if args.filter_indels:
        filter_indels(blocks_by_ids=blocks_by_ids, all_genomes_as_set=(all_genomes if len(args.good_genomes) == 0 else args.good_genomes) - args.bad_genomes)
    if args.filter_duplications:
        filter_duplications(blocks_by_ids=blocks_by_ids)
    all_filtered_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.genomes_order != "":
        genomes_order = args.genomes_order.split(",")
    else:
        genomes_order = sorted(all_filtered_genomes)
    for genome in genomes_order:
        if genome not in all_filtered_genomes:
            logger.critical("Genome \"{genome}\" was specified with teh --genomes-order flag but was not found in the set of filtered genomes {{{filtered_genomes}}}"
                            "".format(genome=genome, filtered_genomes=",".join(sorted(all_filtered_genomes))))
            exit(1)
    if args.ann_genomes == "":
        args.ann_genomes = genomes_order
    else:
        args.ann_genomes = args.ann_genomes.split(",")
        for genome in args.ann_genomes:
            if genome not in genomes_order:
                logger.critical("Annotation genome {ann_genome} was not present in reference genomes {ref_genomes}"
                                "".format(ann_genome=genome, ref_genomes=",".join(sorted(all_filtered_genomes))))
                exit(1)
    sequences = []
    for block_id in sorted(blocks_by_ids.keys()):
        blocks_by_id = blocks_by_ids[block_id]
        blocks = []
        for genome in genomes_order:
            blocks.extend([block for block in blocks_by_id if block.parent_seq.genome_name == genome])
        annotations = []
        for ann_genome in args.ann_genomes:
            for block in blocks:
                if block.parent_seq.genome_name == ann_genome:
                    annotations.append(block.annotation_name)
        annotation = args.ann_delimiter.join(annotations)
        for block in blocks:
            sequences.append(Sequence(name=block.name, parent_seq_id=block.parent_seq.seq_name, start=block.start, end=block.end, strand=block.strand, annotation=annotation, seq_group_id=block.parent_seq.genome_name))
    write_seqi(sequences=sequences, destination=args.output, output_setup=args.o_format, delimiter=args.o_delimiter)
    logger.info("All done!")
    end_time = datetime.datetime.now()
    logger.info("Elapsed time: {el_time}".format(el_time=str(end_time - start_time)))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import logging
import os
import sys
from collections import defaultdict

import configargparse

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
import camsa.utils.ragout.io as ragout_io
from camsa.core.data_structures import AssemblyPoint
import camsa.core.io as camsa_io
from camsa.utils.ragout.shared import filter_indels, filter_duplications
from camsa.utils.ragout.shared import filter_blocks_by_good_genomes, filter_blocks_by_bad_genomes, get_all_genomes_from_blocks


def get_assembly_points(seq_of_blocks):
    result = []
    for l_block, r_block in zip(seq_of_blocks[:-1], seq_of_blocks[1:]):
        ap = AssemblyPoint(seq1=l_block.name, seq2=r_block.name, seq1_or=l_block.strand, seq2_or=r_block.strand, sources={l_block.parent_seq.genome_name, r_block.parent_seq.genome_name}, gap_size=r_block.start - l_block.end)
        result.append(ap)
    return result


if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting Ragout formatted blocks into CAMSA assembly points.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "logging.ini"),
                                                            os.path.join(camsa.root_dir, "utils", "ragout", "ragout_coords2camsa_points.ini")])
    parser.add_argument("-c", "--config", is_config_file=True, help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("ragout_coords", type=str, help="A path to ragout coords file")
    parser.add_argument("--filter-indels", action="store_true", dest="filter_indels", default=False,
                        help="A flag to indicate whether to filter out all synteny blocks, that are not present exactly once in each of the good (all-bad) genomes.")
    parser.add_argument("--no-filter-indels", action="store_false", dest="filter_indels", default=False,
                        help="A flag to indicate whether to filter out all synteny blocks, that are not present exactly once in each of the good (all-bad) genomes.")
    parser.add_argument("--filter-duplications", action="store_true", dest="filter_duplications", default=True,
                        help="A flag to indicate whether to filter out all synteny blocks, that are present more than once in at least on of the good (all-bad) genomes.")
    parser.add_argument("--no-filter-duplications", action="store_false", dest="filter_duplications", default=True,
                        help="A flag to indicate whether to filter out all synteny blocks, that are present more than once in at least on of the good (all-bad) genomes.")
    parser.add_argument("--good-genomes", type=str, default="", help="A coma separated list of genome names, to be processed and conversed.\nDEFAULT: \"\" (i.e., all genomes are good)")
    parser.add_argument("--bad-genomes", type=str, default="Anc0", help="A coma separated list of genome names, to be excluded from processing and conversion.\nDEFAULT: \"\" (i.e., no genomes are bad)")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout)
    parser.add_argument("--o-genomes", type=str, dest="output_genomes", default="",
                        help="A coma separated list of genome names, which will determine the order inferred assembly points to be output.\nDEFAULT: \"\" (i.e., sorted list of good (all - bad) genomes)")
    parser.add_argument("--o-format", type=str, help="")
    parser.add_argument("--o-delimiter", default="\t", type=str, help="")
    parser.add_argument("--c-logging-level", dest="c_logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()

    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.ragout_coords2camsa_points")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    sequences_by_ids, blocks_by_ids = ragout_io.read_from_file(path=args.ragout_coords, silent_fail=False, delimiter="\t")
    all_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.good_genomes != "":
        args.good_genomes = set(args.good_genomes.split(","))
        filter_blocks_by_good_genomes(blocks_by_ids=blocks_by_ids, good_genomes=args.good_genomes)
    if args.bad_genomes != "":
        args.bad_genomes = set(args.bad_genomes.split(","))
        filter_blocks_by_bad_genomes(blocks_by_ids=blocks_by_ids, bad_genomes=args.bad_genomes)
    if args.filter_indels:
        filter_indels(blocks_by_ids=blocks_by_ids, all_genomes_as_set=(all_genomes if len(args.good_genomes) == 0 else args.good_genomes) - args.bad_genomes)
    if args.filter_duplications:
        filter_duplications(blocks_by_ids=blocks_by_ids)
    all_filtered_genomes = get_all_genomes_from_blocks(blocks_as_ids=blocks_by_ids)
    if args.output_genomes == "":
        output_genomes = sorted(all_filtered_genomes)
    else:
        output_genomes = args.output_genomes.split(",")
    for genome in output_genomes:
        if genome not in all_filtered_genomes:
            logger.critical("Genome {genome_name} specified with the --o-genomes flag was not present in all filtered genome {filtered_genomes}"
                            "".format(genome_name=genome, filtered_genomes=",".join(all_filtered_genomes)))
            exit(1)

    aps = []
    for genome in output_genomes:
        blocks_to_convert = []
        for block_id in sorted(blocks_by_ids.keys()):
            blocks = blocks_by_ids[block_id]
            blocks_by_genome = [block for block in blocks if block.parent_seq.genome_name == genome]
            blocks_to_convert.extend(blocks_by_genome)

        blocks_by_seq_ids = defaultdict(list)
        for block in blocks_to_convert:
            blocks_by_seq_ids[block.parent_seq.seq_name].append(block)

        for seq_id in list(blocks_by_seq_ids.keys()):
            blocks_by_seq_ids[seq_id] = sorted(blocks_by_seq_ids[seq_id], key=lambda block: (block.start, block.end))

        for seq_id in list(blocks_by_seq_ids.keys()):
            seq_of_blocks = blocks_by_seq_ids[seq_id]
            seq_aps = get_assembly_points(seq_of_blocks=seq_of_blocks)
            aps.extend(seq_aps)
    logger.info("Writing output to file \"{file_name}\"".format(file_name=args.output))
    camsa_io.write_assembly_points(assembly_points=aps, destination=args.output, output_setup=args.o_format, delimiter=args.o_delimiter)
    logger.info("Elapsed time: {el_time}".format(el_time=str(datetime.datetime.now() - start_time)))



# -*- coding: utf-8 -*-
import csv
import os
import shutil
from collections import defaultdict
from enum import Enum

from camsa.core.data_structures import AssemblyPoint, APFieldOutExtractorConverter, Sequence


def get_fn_relations_for_column_names(fieldnames, aliases):
    canonical_field_names = [aliases.get(name.lower(), name) for name in fieldnames]
    fn_relations = {value: key for key, value in aliases.items()}
    fn_relations.update({canonical: name for name, canonical in zip(fieldnames, canonical_field_names)})
    return fn_relations


PAIRS_COLUMN_ALIASES = {
    ########################
    "species": "origin",
    "organisms": "origin",
    ########################
    "ctg1": "seq1",
    ########################
    "ctg2": "seq2",
    ########################
    "orientation_ctg1": "seq1_or",
    "ctg1_orientation": "seq1_or",
    "ctg1_or": "seq1_or",
    ########################
    "ctg2_orientation": "seq2_or",
    "orientation_ctg2": "seq2_or",
    "ctg2_or": "seq2_or",
    ########################
    "distance": "gap_size",
    "ctg1-ctg2_gap": "gap_size",
    ########################
    "score": "cw",
}


def extract_nullable_value(field, row, fn_relations, default="?"):
    if field not in fn_relations:
        return default
    return row.get(fn_relations[field], default)


def extract_nullable_numerical_value(field, row, fn_relations, default="?"):
    value = extract_nullable_value(field=field, row=row, fn_relations=fn_relations, default=default)
    try:
        value = float(value)
    except (ValueError, TypeError):
        value = default
    return value


def read_pairs(source, delimiter="\t", destination=None, default_cw_eae=1, default_cw_cae=0.9):
    """

    :param source: file like object tot APs data from
    :param delimiter: tab/comma/etc separator
    :param destination: data structure, where information about APs will be stored
    :param default_cw_eae: confidence weight for exact AE, in case ? is provided in source
    :param default_cw_cae: confidence wight for candidate AE, in case ? is provided in source
    :return: destination data structure, that can be viewed as a default dict of list of APs, where key is the source of the AP
    """


    if destination is None:
        destination = defaultdict(list)
    reader = csv.DictReader(source, delimiter=delimiter)
    fieldnames = reader.fieldnames
    fn_relations = get_fn_relations_for_column_names(fieldnames=fieldnames, aliases=PAIRS_COLUMN_ALIASES)
    for row in filter(lambda entry: not entry[fieldnames[0]].startswith("#"), reader):
        origin = row[fn_relations["origin"]]
        cw = extract_nullable_numerical_value(field="cw", row=row, fn_relations=fn_relations)
        if cw == "?":
            cw = default_cw_eae if "?" not in [row[fn_relations["seq1_or"]], row[fn_relations["seq2_or"]]] else default_cw_cae
        gap_size = extract_nullable_numerical_value(field="gap_size", row=row, fn_relations=fn_relations)
        seq1 = row[fn_relations["seq1"]]
        seq2 = row[fn_relations["seq2"]]
        if seq1 == seq2:
            # no support for duplicated seqs present
            continue
        destination[origin].append(AssemblyPoint(seq1=seq1,
                                                 seq2=seq2,
                                                 seq1_or=row[fn_relations["seq1_or"]],
                                                 seq2_or=row[fn_relations["seq2_or"]],
                                                 sources=[row[fn_relations["origin"]]],
                                                 cw=cw,
                                                 gap_size=gap_size))
    return destination


def read_assembly_points_from_input_sources(sources, delimiter="\t", default_cw_eae=1, default_cw_cae=0.75):
    """

    :param sources: list of file paths with input AP data
    :param delimiter: tab/comma/etc separator
    :param default_cw_eae: confidence weight for exact AE, in case ? is provided in source
    :param default_cw_cae: confidence wight for candidate AE, in case ? is provided in source
    :return: destination data structure, that can be viewed as a default dict of list of APs, where key is the source of the AP
    """
    result = defaultdict(list)
    for file_name in sources:
        file_name = os.path.abspath(os.path.expanduser(file_name))
        with open(file_name, "rt") as source:
            read_pairs(source=source, delimiter=delimiter, destination=result,
                       default_cw_eae=default_cw_eae, default_cw_cae=default_cw_cae)
    return result


LENGTHS_COLUMN_ALIASES = {
    ########################
    "ctg_id": "seq_id",
    ########################
    "ctg_length": "seq_length"
}


def read_lengths(source, delimiter="\t", destination=None):
    if destination is None:
        destination = {}
    reader = csv.DictReader(source, delimiter=delimiter)
    fieldnames = reader.fieldnames
    fn_relations = get_fn_relations_for_column_names(fieldnames=fieldnames, aliases=LENGTHS_COLUMN_ALIASES)
    for row in filter(lambda entry: not entry[fieldnames[0]].startswith("#"), reader):
        seq_id = row[fn_relations["seq_id"]]
        seq_length = int(float(row[fn_relations["seq_length"]]))
        destination[seq_id] = Sequence(name=seq_id, length=seq_length)
    return destination


class OrientationChoice(Enum):
    original = 0
    merged = 1


def get_header_and_extract_list(settings):
    data = settings.split("|")
    converters_setups = []
    for entry in data:
        converters_setups.append(entry.split(","))
    header = [entry[0] for entry in converters_setups]
    converters = [APFieldOutExtractorConverter(field_name=entry[1], converter_name=entry[2]) for entry in converters_setups]
    return header, converters


def write_assembly_points(assembly_points, destination, output_setup, delimiter="\t"):
    """ Output a collection of assembly point in a text format to the specified stream

    :param assembly_points: an iterable with a collection of assembly points to be written down
    :param destination: a file like object to write to
    :param delimiter: a separator used for the SV format
    :param orientation_type: a choice for AP relative seq orientations to be displayed (original = input vs inferred = merged).
        Makes a difference only for the un/semi-oriented APs
    """
    intra_delimiter = "," if delimiter != "," else ";"

    writer = csv.writer(destination, delimiter=delimiter)
    header, field_processor = get_header_and_extract_list(settings=output_setup)
    writer.writerow(header)
    for ap in assembly_points:
        assembly_points_entries_list = [processor.extract_field_value_str(ap) for processor in field_processor]
        writer.writerow(assembly_points_entries_list)


def write_seqi(sequences, destination, output_setup, delimiter="\t"):
    writer = csv.writer(destination, delimiter=delimiter)
    header, field_processor = get_header_and_extract_list(settings=output_setup)
    writer.writerow(header)
    for seq in sequences:
        seq_entries_list = [processor.extract_field_value_str(seq) for processor in field_processor]
        writer.writerow(seq_entries_list)


def read_seqi_from_input_sources(source, delimiter="\t", destination=None):
    if destination is None:
        destination = defaultdict(list)
    reader = csv.DictReader(source, delimiter=delimiter)
    fieldnames = reader.fieldnames
    fn_relations = get_fn_relations_for_column_names(fieldnames=fieldnames, aliases=LENGTHS_COLUMN_ALIASES)
    for row in filter(lambda entry: not entry[fieldnames[0]].startswith("#"), reader):
        seq_id = row[fn_relations["seq_id"]]
        length = extract_nullable_numerical_value(field="length", row=row, fn_relations=fn_relations, default=None)
        parent_seq_id = extract_nullable_value(field="parent_seq_id", row=row, fn_relations=fn_relations, default=None)
        start = extract_nullable_numerical_value(field="start", row=row, fn_relations=fn_relations, default=None)
        end = extract_nullable_numerical_value(field="end", row=row, fn_relations=fn_relations, default=None)
        strand = extract_nullable_value(field="strand", row=row, fn_relations=fn_relations, default="+")
        annotation = extract_nullable_value(field="annotation", row=row, fn_relations=fn_relations, default=None)
        seq_group_id = extract_nullable_value(field="seq_group_id", row=row, fn_relations=fn_relations, default="Default")
        seq = Sequence(name=seq_id, length=length, parent_seq_id=parent_seq_id, start=start, end=end, strand=strand, annotation=annotation, seq_group_id=seq_group_id)
        destination[seq.name].append(seq)
    return destination


def remove_dir(dir_path):
    if os.path.exists(dir_path) and os.path.isdir(dir_path):
        shutil.rmtree(dir_path)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import csv
import datetime
import logging
import os
import subprocess
import sys
from collections import defaultdict, deque
from os.path import isfile

import configargparse

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa


class CoordsEntry(object):
    __slots__ = ['scaffold_name', 'fragment_name', 'scaffold_start', 'scaffold_end',
                 'fragment_start', 'fragment_end', 'fragment_length', 'fragment_coverage']

    def __init__(self, scaffold_name, fragment_name, scaf_start, scaf_end, frag_start, frag_end, frag_length, frag_cov):
        self.scaffold_name = scaffold_name
        self.fragment_name = fragment_name
        self.scaffold_start = scaf_start
        self.scaffold_end = scaf_end
        self.fragment_start = frag_start
        self.fragment_end = frag_end
        self.fragment_length = frag_length
        self.fragment_coverage = frag_cov

    @property
    def fragment_orientation(self):
        return "+" if self.fragment_start < self.fragment_end else "-"

    @property
    def mid_coordinate(self):
        return (self.scaffold_start + self.scaffold_end) / 2


class CoordsParser(object):
    """
    This is a simple parser, that always anticipates, that entries coordinates would be in two first meta-columns

    entries names would be in the last
    length and coverage of the query entry is anticipated to be in the second and third columns from the end, respectively
    """

    def __init__(self, source, skip_lines=3, has_header=True, lower_frag_cover=90.0, collapse_consecutive=True):
        self.source = source
        self.skip_lines = skip_lines
        self.has_header = has_header
        self.lower_frag_cover = lower_frag_cover
        self.collapse_consecutive = collapse_consecutive

    def parse_data(self):
        result = defaultdict(list)
        for cnt, entry in enumerate(self.source):
            if cnt < self.skip_lines:
                continue
            elif cnt in [self.skip_lines, self.skip_lines + 1] and self.has_header:
                continue
            coords_entry = self.from_string(string=entry)
            # if coords_entry.fragment_coverage < self.lower_frag_cover:
            #     continue
            result[coords_entry.scaffold_name].append(coords_entry)
        for name in result.keys():
            result[name] = sorted(result[name], key=lambda coord_entry: coords_entry.mid_coordinate)
        if self.collapse_consecutive:
            for name in list(result.keys()):
                entries = result[name]
                new_entries = []
                current_entry = entries[0]
                for entry in entries[1:]:
                    if entry.fragment_name == current_entry.fragment_name \
                            and entry.fragment_orientation == current_entry.fragment_orientation:
                        current_entry.fragment_coverage += entry.fragment_coverage
                        if entry.fragment_orientation == "+":
                            current_entry.scaffold_end = entry.scaffold_end
                        else:
                            current_entry.scaffold_start = entry.scaffold_start
                    else:
                        new_entries.append(current_entry)
                        current_entry = entry
                new_entries.append(current_entry)
                result[name] = new_entries
        for name in list(result.keys()):
            entries = result[name]
            new_entries = []
            for entry in entries:
                if entry.fragment_coverage >= self.lower_frag_cover:
                    new_entries.append(entry)
            result[name] = new_entries
        return result

    @staticmethod
    def from_string(string):
        data = [entry.strip() for entry in string.split(" | ")]
        reference_coords_data, query_coords_data = data[0], data[1]
        names_data = data[-1]
        lengths_data, coverage_data = data[-3], data[-2]
        ref_start, ref_end = [int(entry.strip()) for entry in reference_coords_data.split()]
        frag_start, frag_end = [int(entry.strip()) for entry in query_coords_data.split()]
        ref_name, frag_name = [entry.strip() for entry in names_data.split()]
        frag_length = float(lengths_data.split()[1].strip())
        frag_cov = float(coverage_data.split()[1].strip())
        return CoordsEntry(scaffold_name=ref_name, fragment_name=frag_name,
                           scaf_start=ref_start, scaf_end=ref_end,
                           frag_start=frag_start, frag_end=frag_end,
                           frag_length=frag_length, frag_cov=frag_cov)


def get_file_prefix(file_name):
    return os.path.basename(os.path.splitext(file_name)[0])


def run_nucmer(contigs_file_name, reference_file_name, nucmer_executable_path, cli_arguments, output_dir, logs_dir, nucmer_stdout=None, nucmer_stderr=None):
    prefix = get_file_prefix(reference_file_name)
    if nucmer_stdout is None:
        nucmer_stdout = os.path.join(logs_dir, "nucmer_{prefix}.stdout.txt".format(prefix=prefix))
    if nucmer_stderr is None:
        nucmer_stderr = os.path.join(logs_dir, "nucmer_{prefix}.stderr.txt".format(prefix=prefix))
    cli_args = [nucmer_executable_path] + cli_arguments.split(" ") + ['-p', os.path.join(output_dir, prefix), reference_file_name, contigs_file_name, '>', nucmer_stdout, '2>', nucmer_stderr]
    logger.info("\t" + " ".join(cli_args))
    exitcode = subprocess.call(" ".join(cli_args), shell=True)
    if exitcode == 0:
        logger.info("NUCmer finished running for \"{scaffolds_file}\" scaffolds file.")
        logger.debug("NUCmer alignment output is stored in \"{delta_output}\" file."
                     "".format(delta_output=os.path.join(output_dir, "{prefix}.delta".format(prefix=prefix))))
        logger.debug("NUCmer log is stored in \"{nucmer_log}\".".format(scaffolds_file=reference_file_name, nucmer_log=nucmer_stdout))
    else:
        logger.error("NUCmer exited with non-zero code, running for \"{scaffolds}\" scaffolds file.".format(scaffolds=reference_file_name))
        logger.error("NUCmer logs are stored in:")
        logger.error("\tstdout: \"{nucmer_log}\"".format(scaffolds=reference_file_name, nucmer_log=nucmer_stdout, nucmer_err=nucmer_stderr))
        logger.error("\tstderr: \"{nucmer_err}\"".format(scaffolds=reference_file_name, nucmer_log=nucmer_stdout, nucmer_err=nucmer_stderr))
    return exitcode


def run_show_coords(delta_file_name, output_dir, logs_dir, show_coords_executable_path, cli_arguments, show_coords_stderr=None):
    prefix = get_file_prefix(delta_file_name)
    if prefix.endswith(".filtered"):
        prefix = prefix[:-9]
    output_file_name = os.path.join(output_dir, prefix + ".coords")
    if show_coords_stderr is None:
        show_coords_stderr = os.path.join(logs_dir, "show_coords_{prefix}.stderr.txt".format(prefix=prefix))
    cli_args = [show_coords_executable_path] + cli_arguments.split(" ") + [delta_file_name, '>', output_file_name, '2>', show_coords_stderr]
    logger.info("\t" + " ".join(cli_args))
    exitcode = subprocess.call(" ".join(cli_args), shell=True)
    if exitcode == 0:
        logger.info("show-coords util has finished running for \"{delta_file}\".".format(delta_file=delta_file_name))
        logger.debug("show-coords output is stored in \"{coords_output}\" file.".format(coords_output=output_file_name))
    else:
        logger.error("show-coords exited with non-zero code, running for \"{delta_file}\".".format(delta_file=delta_file_name, show_coords_error_log=show_coords_stderr))
        logger.error("\tshow-coords error log is stored in \"{show_coords_error_log}\"".format(delta_file=delta_file_name, show_coords_error_log=show_coords_stderr))
    return exitcode


def run_delta_filter(delta_file_name, output_dir, logs_dir, delta_filter_executable_path, cli_arguments, delta_filter_stderr=None):
    prefix = get_file_prefix(delta_file_name)
    output_file_name = os.path.join(output_dir, prefix + ".filtered.delta")
    if delta_filter_stderr is None:
        delta_filter_stderr = os.path.join(logs_dir, "delta_filter_{prefix}.stderr.txt".format(prefix=prefix))
    cli_args = [delta_filter_executable_path] + cli_arguments.split(" ") + [delta_file_name, ">", output_file_name, "2>", delta_filter_stderr]
    logger.info("\t" + " ".join(cli_args))
    exitcode = subprocess.call(" ".join(cli_args), shell=True)
    if exitcode == 0:
        logger.info("delta-filter util has finished running for \"{delta_file}\"".format(delta_file=delta_file_name))
        logger.debug("delta-filter output is stored in \"{delta_filter_output}\" file".format(delta_filter_output=output_file_name))
    else:
        logger.error("delta-filter exited with non-zero code, running for \"{delta_file}\"".format(delta_file=delta_file_name, show_coords_error_log=delta_filter_stderr))
        logger.error("\tdelta-filter error log is stored in \"{delta_filter_error_log}\"".format(delta_file=delta_file_name, delta_filter_error_log=delta_filter_stderr))
    return exitcode


def exit_program():
    logger.critical("An error was encountered and `--ensure-all` flag was set to true. fasta2camsa_points.py exists.")
    exit(1)


def get_assembly_points_from_aligned_contigs(coords_entries, strategy):
    if strategy == "mid-point-sort":
        entries = sorted(coords_entries, key=lambda it: it.mid_coordinate)
        return zip(entries[:-1], entries[1:])
    if strategy == "sliding-window":
        to_go = deque(sorted(coords_entries, key=lambda it: (it.scaffold_start, it.scaffold_end)))
        processing = set()
        finished = []
        pairs = []
        for current in to_go:
            to_be_finished = sorted([entry for entry in processing if entry.scaffold_end <= current.scaffold_start], key=lambda i: i.scaffold_end)
            if len(to_be_finished) == 0 and len(finished) != 0:
                pairs.append((finished[-1], current))
            for item in to_be_finished:
                pairs.append((item, current))
                processing.remove(item)
                finished.append(item)
            processing.add(current)
        return pairs


def filter_fully_covered_contigs(contigs):
    start_entries = [(contig.scaffold_start, contig, "start") if contig.fragment_orientation == "+" else (contig.scaffold_end, contig, "start") for contig in contigs]
    end_entries = [(contig.scaffold_end, contig, "end") if contig.fragment_orientation == "-" else (contig.scaffold_start, contig, "end") for contig in contigs]
    entries = sorted(start_entries + end_entries, key=lambda it: it[0])
    result_chain = []
    processing = deque()
    if "3600" in {contig.fragment_name for contig in contigs}:
        a = 5
    for entry in entries:
        if entry[2] == "start":
            processing.append(entry[1])
        else:
            if entry[1] == processing[0]:
                result_chain.append(entry[1])
                processing.popleft()
            else:
                processing.remove(entry[1])
    return result_chain


if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting FASTA formatted scaffolding results for further CAMSA processing.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "fasta", "fasta2camsa_points.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])

    parser.add_argument("contigs", metavar="CONTIGS", help="fasta formatted file with contigs, that served as input for scaffolding purposes")
    parser.add_argument("scaffolds", metavar="SCAFFOLDS", nargs="+", help="fasta formatted result files of contigs scaffolding")
    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("-c", "--config", is_config_file=True, help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")
    parser.add_argument("-o", "--output-dir", metavar="OUTPUT_DIR", dest="output_dir", default="output", help="Output directory to store temporary and final files.\nDEFAULT: ./output/")
    parser.add_argument("--tmp-dir", metavar="TMP_DIR", dest="tmp_dir", default=None, help="Directory with all intermediate (.delta | .coords) files.")
    parser.add_argument("--overwrite", dest="overwrite", action="store_true", help="Disregards already present \"*.delta\" as well as \"*.coords\" and runs all the stages from scratch.\nDEFAULT: False")
    parser.add_argument("--ensure-all", dest="ensure_all", default=False, type=bool, help="Flag indicating, that is any subprocess of this converter fails, than the whole operation will be canceled.\nDEFAULT: False")

    parser.add_argument("--nucmer-cli-arguments", dest="nucmer_cli_arguments",
                        help="Command line argument, used when running \"nucmer\" software\nDEFAULT: -maxmatch -c 100")

    parser.add_argument("--show-coords-cli-arguments", dest="show_coords_cli_arguments",
                        help="Command line argument, used when running show-coords software\nDEFAULT: -r -c -l")

    parser.add_argument("--delta-filter-cli-arguments", dest="delta_filter_cli_arguments",
                        help="Command line argument, used when running delta-filter software\nDEFAULT: -r -q")

    #######################################################################################################################
    # will definitely make two options this work, once the python bug is fixed: http://bugs.python.org/issue15112
    # by now all the delta files, that match the contigs files name will results in alignment stage for those file to be
    #######################################################################################################################
    # parser.add_argument("--delta-files", default=None, nargs="*", help="paths to the nucmer \"*.delta\" files. For each file present, corresponding alignment stage will be skipped")
    # parser.add_argument("--coords-files", default=None, nargs="*", help="paths to the nucmer \"*.coords\" files. For each file present, corresponding show-coords stage will be skipped")

    parser.add_argument("--nucmer-path", dest="nucmer", help="full path to the nucmer executable.\nDEFAULT: nucmer")
    parser.add_argument("--show-coords-path", dest="show_coords", help="full path to the show-coords executable.\nDEFAULT: show-coords")
    parser.add_argument("--delta-filter-path", dest="delta_filter", help="full path to the delta-filter executable.\nDEFAULT: delta-filter")

    parser.add_argument("--c-cov-threshold", default=90.0, type=float, dest="c_cov_threshold",
                        help="lower coverage bound with respect to each aligned contig. All contigs with coverage less than the threshold are omitted.\nDEAFULT: 90.0")
    parser.add_argument("--c-no-collapse-cons-alignments", default=True, action="store_false", dest="collapse_consecutive_alignments")
    parser.add_argument("--c-keep-fully-covered-contigs", action="store_true", default=False,
                        help="Whether to keep contigs, that are mapped some other contigs, or not.\nDEFAULT: False")
    parser.add_argument("--c-coords-pairs-strategy", choices=["mid-point-sort", "sliding-window", "all"], default="mid-point-sort", dest="coords_to_pairs_strategy",
                        help="A strategy that determines on how assembly pairs from contigs on each scaffold are inferred.\n"
                             "\"mid-point-sort\" -- all contig mapping on each scaffold are sorted by their mid coordinate (start + end) / 2.\n\tSorted sequence of contigs determines n-1 assembly points.\n"
                             "\"sliding-window\" -- all pairs of adjacent extremities of non overlapping contigs will be reported as assembly points."
                             "\nDEFAULT: mid-point-sort ")
    parser.add_argument("--c-logging-level", type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    args = parser.parse_args()
    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.fasta2camsa_pairs")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    args.output_dir = os.path.expanduser(args.output_dir)
    args.output_dir = os.path.abspath(args.output_dir)
    args.tmp_dir = os.path.join(args.output_dir, "fasta2camsa") if args.tmp_dir is None else os.path.abspath(os.path.expanduser(args.tmp_dir))
    args.logs_dir = os.path.join(args.tmp_dir, "logs")

    if not os.path.exists(args.output_dir):
        logger.debug("Output directory \"{directory}\" doesn't exists. Creating one.".format(directory=args.output_dir))
        os.makedirs(args.output_dir)
    if not os.path.exists(args.tmp_dir):
        logger.debug("Creating \"tmp\" directory, where intermediate tools results will be stored")
        os.makedirs(args.tmp_dir)
    if not os.path.exists(args.logs_dir):
        logger.debug("Creating log directory, that would contain all the external tools logs information")
        os.makedirs(args.logs_dir)

    files_in_tmp_dir = [f for f in os.listdir(args.tmp_dir) if isfile(os.path.join(args.tmp_dir, f))]
    delta_files_base_names = {os.path.splitext(f)[0] for f in files_in_tmp_dir if os.path.splitext(f)[1] == ".delta"}
    coords_files_base_names = {os.path.splitext(f)[0] for f in files_in_tmp_dir if os.path.splitext(f)[1] == ".coords"}

    for scaffolds_file in args.scaffolds:
        prefix = get_file_prefix(file_name=scaffolds_file)
        logger.info("Working with \"{scaffolds_file}\"".format(scaffolds_file=scaffolds_file))
        ##################################################################################################
        # obtaining NUCmer alignments results in a form of *.delta file.
        ##################################################################################################
        if (prefix in delta_files_base_names or prefix in coords_files_base_names) and not args.overwrite:
            logger.info("File \"{prefix}.delta\" or \"{prefix}.coords\" is present in the tmp folder. "
                        "Skipping alignment [show-coords] stage for \"{scaffold_file}\" scaffolds file."
                        "".format(prefix=prefix, scaffold_file=scaffolds_file))
        else:
            logger.info("Running NUCmer for \"{scaffolds_file}\" scaffolds file, using \"{contigs_file}\" as query. This might take time."
                        "".format(contigs_file=args.contigs, scaffolds_file=scaffolds_file))
            logger.debug("Results will be stored in \"{prefix}.delta\"."
                         "".format(prefix=prefix))
            result = run_nucmer(contigs_file_name=args.contigs, reference_file_name=scaffolds_file,
                                nucmer_executable_path=args.nucmer,
                                output_dir=args.tmp_dir,
                                logs_dir=args.logs_dir,
                                cli_arguments=args.nucmer_cli_arguments)
            if result != 0 and args.ensure_all:
                exit_program()

        ##################################################################################################
        # obtaining NUCmer alignment results in a form of *.coord file.
        ##################################################################################################
        if prefix in coords_files_base_names and not args.overwrite:
            logger.info("File \"{prefix}.coords\" is present in the tmp folder."
                        " Skipping show-coords stage for \"{prefix}.delta\" file."
                        "".format(prefix=prefix))
        else:
            delta_file_name = [f for f in os.listdir(args.tmp_dir)]
            delta_file_name = [f for f in delta_file_name if os.path.isfile(os.path.join(args.tmp_dir, f))]
            delta_file_name = [f for f in delta_file_name if os.path.splitext(f)[1] == ".delta"]
            delta_file_name = [f for f in delta_file_name if os.path.splitext(f)[0] == prefix]
            if len(delta_file_name) == 0:
                logger.error("Delta file for prefix=\"{prefix}\" was not found in the output folder.".format(prefix=prefix))
                if args.ensure_all:
                    exit_program()
                continue
            original_name_prefix = ".".join(delta_file_name[0].split(".")[:-1])
            delta_file_name = os.path.join(args.tmp_dir, original_name_prefix + ".delta")
            logger.info("Running delta-filter util for \"{delta_file}\" file.".format(delta_file=delta_file_name))
            logger.debug("Results will be stored in \"{prefix}.filtered.delta\"".format(prefix=prefix))
            result = run_delta_filter(delta_file_name=delta_file_name, output_dir=args.tmp_dir, logs_dir=args.logs_dir,
                                      delta_filter_executable_path=args.delta_filter,
                                      cli_arguments=args.delta_filter_cli_arguments)
            if result != 0 and args.ensure_all:
                exit_program()

            delta_file_name = os.path.join(args.tmp_dir, original_name_prefix + ".filtered.delta")
            logger.info("Running show-coords util for \"{delta_file}\" file.".format(delta_file=delta_file_name))
            logger.debug("Results will be stored in \"{prefix}.coords\"".format(prefix=original_name_prefix))
            result = run_show_coords(delta_file_name=delta_file_name, output_dir=args.tmp_dir, logs_dir=args.logs_dir, show_coords_executable_path=args.show_coords, cli_arguments=args.show_coords_cli_arguments)
            if result != 0 and args.ensure_all:
                exit_program()

        ##################################################################################################
        # translating *.coords language of NUCmer into the language of assembly pairs of CAMSA
        # also generates a fragments' length mapping file
        ##################################################################################################
        coords_file = os.path.join(args.tmp_dir, prefix + ".coords")
        if not os.path.exists(coords_file):
            logger.error("Coords file for \"{prefix}\" doesn't exist in the tmp output directory.".format(prefix=prefix))
            if args.ensure_all:
                exit_program()
            continue
        with open(coords_file, "rt") as source:
            logger.info("Parsing \"{coords_file}\".".format(coords_file=coords_file))
            parser = CoordsParser(source=source, lower_frag_cover=args.c_cov_threshold, collapse_consecutive=args.collapse_consecutive_alignments)
            chains = parser.parse_data()

            result_file = os.path.join(args.output_dir, prefix + ".camsa.points")
            with open(result_file, "wt") as destination:
                logger.info("Writing coords data in terms of CAMSA assembly points in \"{camsa_input_file}\"".format(camsa_input_file=result_file))
                writer = csv.writer(destination, delimiter="\t")
                writer.writerow(['origin', 'seq1', 'seq1_or', 'seq2', 'seq2_or', 'gap_size', 'cw'])
                for scaffold, contigs_chain in chains.items():
                    if not args.c_keep_fully_covered_contigs:
                        contigs_chain = filter_fully_covered_contigs(contigs=contigs_chain)
                    assembly_points = get_assembly_points_from_aligned_contigs(coords_entries=contigs_chain,
                                                                               strategy=args.coords_to_pairs_strategy)
                    for left, right in assembly_points:
                        writer.writerow([prefix,
                                         left.fragment_name, left.fragment_orientation,
                                         right.fragment_name, right.fragment_orientation,
                                         min(right.scaffold_start, right.scaffold_end) - max(left.scaffold_end, left.scaffold_start),
                                         "?"])
            logger.info("Finished converting data for \"{prefix}\"".format(prefix=prefix))
    logger.info("Elapsed time: {el_time}".format(el_time=str(datetime.datetime.now() - start_time)))



# -*- coding: utf-8 -*-
import os
import re

import sys
from collections import defaultdict

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from camsa.utils.ragout.data_structures import RagoutSequence, Block

block_id_pattern = re.compile("Block #(?P<block_id>\d+)")
ENTRIES_SEPARATOR = "--------------------------------------------------------------------------------"


def read_from_file(path, silent_fail=False, delimiter="\t"):
    with open(path, "rt") as source:
        whole_text = source.read()
        sequences_by_ids, blocks_by_ids = parse_ragout_coords(ragout_coords_as_a_string=whole_text, delimiter=delimiter, silent_fail=silent_fail)
        return sequences_by_ids, blocks_by_ids


def parse_ragout_coords(ragout_coords_as_a_string, delimiter="\t", silent_fail=False):
    data = [string_entry.strip() for string_entry in ragout_coords_as_a_string.split(ENTRIES_SEPARATOR) if len(string_entry.strip()) > 0]
    seq_part = data[0]
    blocks = data[1:]
    sequences_by_ids = parse_ragout_seq_part(entry_as_a_string=seq_part, delimiter=delimiter, silent_fail=silent_fail)
    all_blocks = []
    for block_string in blocks:
        current_blocks = parse_ragout_block_entry(entry_as_a_string=block_string, sequences_by_ids=sequences_by_ids, silent_fail=silent_fail, delimiter=delimiter)
        all_blocks.extend(current_blocks)
    blocks_by_ids = defaultdict(list)
    for block in all_blocks:
        blocks_by_ids[block.name].append(block)
    return sequences_by_ids, blocks_by_ids


def get_seq_headers_columns_indexes(headers_string, delimiter, silent_fail=False):
    """
            Seq_id	Size	Description
    """
    headers_string = headers_string.strip()
    data = headers_string.split(delimiter)
    seq_id_columns = data.index("Seq_id")
    length_column = data.index("Size")
    description_column = data.index("Description")
    if any(map(lambda x: x < 0, [seq_id_columns, length_column, description_column])) and not silent_fail:
        raise ValueError("Could not parse block headers")
    return {"seq_id_column": seq_id_columns, "seq_length_column": length_column, "seq_description_column": description_column}


def parse_ragout_seq_part(entry_as_a_string, delimiter="\t", silent_fail=False):
    entries = [entry.strip() for entry in entry_as_a_string.split("\n")]
    seq_headers = entries[0]
    seq_entries = entries[1:]
    columns_indexes = get_seq_headers_columns_indexes(headers_string=seq_headers, delimiter=delimiter, silent_fail=silent_fail)
    if any(map(lambda x: x < 0, columns_indexes.values())) and silent_fail:
        return
    result_sequences = {}
    for entry_string in seq_entries:
        seq = parse_ragout_seq_entry(entry_as_a_string=entry_string,
                                     seq_id_column=columns_indexes["seq_id_column"],
                                     seq_length_column=columns_indexes["seq_length_column"],
                                     seq_description_column=columns_indexes["seq_description_column"], delimiter=delimiter)
        result_sequences[seq.ragout_id] = seq
    return result_sequences


def parse_ragout_seq_entry(entry_as_a_string, seq_id_column=0, seq_length_column=1, seq_description_column=2, delimiter="\t"):
    """
        Seq_id	Size	Description
        27	11234	Anc0.Anc0refChr7043
    """
    data = entry_as_a_string.split(delimiter)
    seq_id = int(float(data[seq_id_column]))
    seq_length = int(float(data[seq_length_column]))
    seq_name = data[seq_description_column]
    return RagoutSequence(name=seq_name, ragout_id=seq_id, length=seq_length)


def parse_ragout_block_entry(entry_as_a_string, sequences_by_ids=None, silent_fail=False, delimiter="\t"):
    """
        Block #4873
        Seq_id	Strand	Start	End	Length
        18386	+	36450191	36456417	6226
        19238	-	6907	699	6208
        24172	-	9270	2915	6355
        28652	-	606707	600419	6288
        30888	-	606707	600419	6288
        ....
    """
    entries = [entry.strip() for entry in entry_as_a_string.split("\n")]
    block_id_string = entries[0]
    columns_headers = entries[1]
    blocks = entries[2:]
    block_id = parse_block_id_string(data_string=block_id_string, silent_fail=silent_fail)
    if block_id is None and silent_fail:
        return
    columns_indexes = get_blocks_headers_columns_indexes(headers_string=columns_headers, delimiter=delimiter, silent_fail=silent_fail)
    if any(map(lambda x: x < 0, columns_indexes.values())) and silent_fail:
        return
    result_blocks = []
    for block_string in blocks:
        result_blocks.append(parse_ragout_block_string(data_string=block_string, block_id=block_id,
                                                       seq_id_column=columns_indexes["seq_id_column"],
                                                       strand_column=columns_indexes["strand_column"],
                                                       start_column=columns_indexes["start_column"],
                                                       end_column=columns_indexes["end_column"]))
    if sequences_by_ids is not None:
        for block in result_blocks:
            try:
                block.parent_seq = sequences_by_ids[block.parent_seq]
            except KeyError:
                if not silent_fail:
                    raise
    return result_blocks


def get_blocks_headers_columns_indexes(headers_string, delimiter="\t", silent_fail=False):
    """
        Seq_id	Strand	Start	End	Length
    """
    headers_string = headers_string.strip()
    data = headers_string.split(delimiter)
    seq_id_columns = data.index("Seq_id")
    strand_column = data.index("Strand")
    start_column = data.index("Start")
    end_column = data.index("End")
    if any(map(lambda x: x < 0, [seq_id_columns, strand_column, start_column, end_column])) and not silent_fail:
        raise ValueError("Could not parse block headers")
    return {"seq_id_column": seq_id_columns, "strand_column": strand_column, "start_column": start_column, "end_column": end_column}


def parse_block_id_string(data_string, pattern=block_id_pattern, silent_fail=False):
    """
        Block #xxxx
    """
    data_string = data_string.strip()
    match = pattern.match(data_string)
    if match is None:
        if not silent_fail:
            raise ValueError("Block id string was not successfully matched by the regex pattern")
    else:
        block_id = int(match.group("block_id"))
        return block_id


def parse_ragout_block_string(data_string, block_id, seq_id_column=0, strand_column=1, start_column=2, end_column=3, delimiter="\t"):
    """
        Seq_id	Strand	Start	End	Length
        18386	+	36450191	36456417	6226
    """
    data_string = data_string.strip()
    data = data_string.split(delimiter)
    seq_id = int(data[seq_id_column])
    strand = data[strand_column]
    start = int(data[start_column])
    end = int(data[end_column])
    if strand == "-":
        start, end = end, start
    block = Block(name=block_id, start=start, end=end, strand=strand, parent_seq=seq_id)
    return block



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import csv
import datetime
import logging
import os
import sys

import configargparse
from Bio import SeqIO


sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
from camsa.core.data_structures import Sequence
from camsa.core.io import write_seqi

if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting FASTA formatted fragments further CAMSA processing.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "fasta", "fasta2camsa_seqi.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])

    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("-o", "--output-file", metavar="OUTPUT", dest="output", type=configargparse.FileType("wt"), default=sys.stdout,
                        help="A file to which the CAMSA readable fragments lengths would be written. Standard extension is \".camsa.lengths\".\nDEFAULT: stdout")
    parser.add_argument("--o-format", type=str, help="")
    parser.add_argument("contigs", nargs="+", metavar="CONTIGS", type=configargparse.FileType("rt"), default=sys.stdin,
                        help="A list of input *.fasta files with contigs.\nDEFAULT: stdin")
    parser.add_argument("--c-logging-level", dest="logging_level", default=logging.INFO, type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")
    parser.add_argument("--o-delimiter", type=str, default="\t",
                        help="A single character string, used as a delimiter in the output (t)/(c)sv file.\nDEFAULT: \\t")
    args = parser.parse_args()
    start_time = datetime.datetime.now()

    logger = logging.getLogger("CAMSA.utils.fasta2camsa_seqi")
    ch = logging.StreamHandler()
    ch.setLevel(args.logging_level)
    logger.setLevel(args.logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    entries = {}
    for f in args.contigs:
        try:
            seq_group_id = os.path.splitext(os.path.basename(f.name))[0]
        except (ValueError, TypeError, AttributeError):
            seq_group_id = None
        logger.info("Processing file: \"{file_name}\"".format(file_name=f))
        cnt = 0
        for record in SeqIO.parse(f, "fasta"):
            seq = Sequence(name=record.id, parent_seq_id=record.id, start=0, end=len(record.seq), strand="+", annotation=record.description, seq_group_id=seq_group_id)
            entries[record.id] = seq
            cnt += 1
        logger.info("Processed {cnt} fasta records".format(cnt=cnt))
    sequences = sorted(entries.values(), key=lambda seq: seq.name)
    write_seqi(sequences=sequences, destination=args.output, output_setup=args.o_format, delimiter=args.o_delimiter)
    logger.info("All done!")
    end_time = datetime.datetime.now()
    logger.info("Elapsed time: {el_time}".format(el_time=str(end_time - start_time)))



# -*- coding: utf-8 -*-

from camsa.core.data_structures import Sequence


class Block(object):
    def __init__(self, name, start, end, strand, parent_seq=None, annotation=None):
        self.name = name
        self.start = start
        self.end = end
        self.strand = strand
        self.parent_seq = parent_seq
        self._annotation = annotation

    @property
    def length(self):
        return self.end - self.start

    @property
    def annotation_name(self):
        if self._annotation is None:
            return ".".join([str(self.parent_seq.genome_name), str(self.parent_seq.seq_name), str(self.name), str(self.start) + "-" + str(self.end), str(self.length)])
        return self._annotation

    def __str__(self):
        return self.annotation_name

    def __repr__(self):
        return str(self)


class RagoutSequence(Sequence):
    def __init__(self, name, ragout_id, length=None):
        super(RagoutSequence, self).__init__(name=name, length=length)
        self.ragout_id = ragout_id

    @property
    def genome_name(self):
        return self.name.split(".")[0]

    @property
    def seq_name(self):
        return self.name.split(".", 1)[1]





# -*- coding: utf-8 -*-
import logging
import os

VERSION = "1.1.0b9"
full_description_template = "| {names:<76} |\n| {affiliations:<76} |\n| {dummy:<76} |\n| {tool:<76} |\n| {dummy:<76} |\n| {information:<76} |\n| {contact:<76} |\n"
CAMSA_AUTHORS = "Sergey Aganezov & Max A. Alekseyev (c)"
AFFILIATIONS = "Computational Biology Institute, The George Washington University"
CONTACT = "With any questions, please, contact Sergey Aganezov [aganezov(at)gwu.edu]"
CAMSA_DOCS_URL = "github.com/compbiol/camsa/wiki"
formatter = logging.Formatter('%(asctime)s - %(name)-15s - %(levelname)-7s - %(message)s')
root_dir = os.path.dirname(os.path.abspath(__file__))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import enum
import logging
import os
import sys
from collections import defaultdict

import configargparse

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

import camsa
import camsa.core.io as camsa_io
from camsa.core.data_structures import AssemblyPoint


class Component(object):
    def __init__(self, object_id, object_beg, object_end, part_number, component_type):
        self.object_id = object_id
        self.object_beg = object_beg
        self.object_end = object_end
        self.part_number = part_number
        self.component_type = component_type

    @classmethod
    def from_agp_data(cls, data):
        raise NotImplementedError("Only specific heirs have implementation for this method")

    def is_gap_component(self):
        raise NotImplementedError("Only specific heirs have implementation for this method")

    @staticmethod
    def is_scaffold_component(component_type):
        return component_type not in ["U", "N"]


class ScaffoldComponent(Component):
    def __init__(self, object_id, object_beg, object_end, part_number, component_type, component_id, component_beg, component_end, orientation):
        super(ScaffoldComponent, self).__init__(object_id=object_id, object_beg=object_beg, object_end=object_end, part_number=part_number, component_type=component_type)
        self.component_id = component_id
        self.component_beg = component_beg
        self.component_end = component_end
        self.orientation = orientation

    @classmethod
    def from_agp_data(cls, data):
        object_id = data[0]
        object_beg = int(data[1])
        object_end = int(data[2])
        part_number = int(data[3])
        component_type = data[4]
        component_id = data[5]
        component_beg = int(data[6])
        component_end = int(data[7])
        orientation = data[8]
        return cls(object_id=object_id, object_beg=object_beg, object_end=object_end, part_number=part_number, component_type=component_type,
                   component_id=component_id, component_beg=component_beg, component_end=component_end, orientation=orientation)

    @property
    def is_gap_component(self):
        return False


class GapComponent(Component):
    def __init__(self, object_id, object_beg, object_end, part_number, component_type, gap_length, gap_type, linkage, linkage_evidence):
        super(GapComponent, self).__init__(object_id=object_id, object_beg=object_beg, object_end=object_end, part_number=part_number, component_type=component_type)
        self.gap_length = gap_length
        self.gap_type = gap_type
        self.linkage = linkage
        self.linkage_evidence = linkage_evidence

    @classmethod
    def from_agp_data(cls, data):
        object_id = data[0]
        object_beg = int(data[1])
        object_end = int(data[2])
        part_number = int(data[3])
        component_type = data[4]
        gap_length = int(data[5])
        gap_type = data[6]
        linkage = data[7]
        linkage_evidence = data[8]
        return cls(object_id=object_id, object_beg=object_beg, object_end=object_end, part_number=part_number, component_type=component_type,
                   gap_length=gap_length, gap_type=gap_type, linkage=linkage, linkage_evidence=linkage_evidence)

    @property
    def is_gap_component(self):
        return True


def get_index_of_first_non_gap_element(component_sequence):
    for cnt, component in enumerate(component_sequence):
        if not component.is_gap_component:
            return cnt
    return -1


def get_next_non_gap_element(components, current_index):
    for i in range(current_index + 1, len(components)):
        next = components[i]
        if not next.is_gap_component:
            return i, next
    return None, None


def get_gap_size(components, current_index, next_index):
    if current_index == next_index - 1:
        return 1
    gap_components = components[current_index+1:next_index]
    result = 0
    for component in gap_components:
        assert component.is_gap_component
        if component.component_type == "U":
            return "?"
        result += component.gap_length
    return result


def camsa_orientation_from_agp(agp_or):
    if agp_or in ["?", "0", "na"]:
        return "?"
    return agp_or


def object_as_camsa_points(object_as_components, extra_data):
    result = []
    components = object_as_components
    f_non_gap_element = get_index_of_first_non_gap_element(component_sequence=components)
    current_element = components[f_non_gap_element]
    current_index = f_non_gap_element
    next_index, next_element = get_next_non_gap_element(components=components, current_index=current_index)
    while next_element is not None:
        gap_size = get_gap_size(components, current_index, next_index)
        seq1_or = camsa_orientation_from_agp(agp_or=current_element.orientation)
        seq2_or = camsa_orientation_from_agp(agp_or=next_element.orientation)
        assembly_point = AssemblyPoint(seq1=current_element.component_id, seq2=next_element.component_id,
                                       seq1_or=seq1_or, seq2_or=seq2_or, sources=[extra_data.origin],
                                       gap_size=gap_size, cw="?")
        result.append(assembly_point)
        current_element = next_element
        current_index = next_index
        next_index, next_element = get_next_non_gap_element(components=components, current_index=current_index)
    return result


if __name__ == "__main__":
    full_description = camsa.full_description_template.format(
        names=camsa.CAMSA_AUTHORS,
        affiliations=camsa.AFFILIATIONS,
        dummy=" ",
        tool="Converting AGPv2.0 formatted scaffolding results for further CAMSA processing.",
        information="For more information refer to {docs}".format(docs=camsa.CAMSA_DOCS_URL),
        contact=camsa.CONTACT)
    full_description = "=" * 80 + "\n" + full_description + "=" * 80 + "\n"
    parser = configargparse.ArgParser(description=full_description, formatter_class=configargparse.RawTextHelpFormatter,
                                      default_config_files=[os.path.join(camsa.root_dir, "utils", "agp", "agp2camsa_points.ini"),
                                                            os.path.join(camsa.root_dir, "logging.ini")])
    parser.add_argument("-c", "--config", is_config_file=True,
                        help="Config file overwriting some of the default settings as well as any flag starting with \"--\".")
    parser.add_argument("--version", action="version", version=camsa.VERSION)
    parser.add_argument("--c-logging-level", type=int,
                        choices=[logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL],
                        help="Logging level for the converter.\nDEFAULT: {info}".format(info=logging.INFO))
    parser.add_argument("--c-logging-formatter-entry",
                        help="Format string for python logger.")

    parser.add_argument("agp", type=configargparse.FileType("rt"), nargs="+", default=sys.stdin,
                        help="Input stream of AGPv2 formatted scaffold assemblies\nDEFAULT: stdin")
    parser.add_argument("--origin", type=str, default=None,
                        help="Identifier for the assembly, that would be specified in the \"origin\" column of CAMSA points\nDEFAULT: inferred from input file names")
    parser.add_argument("--o-format", type=str,
                        help="The CAMSA-out formatting for the assembly points obtained form the AGPv2 formatted scaffold assemblies")
    parser.add_argument("-o", "--output", type=configargparse.FileType("wt"), default=sys.stdout,
                        help="The stream where CAMSA formatted assembly points are outputted\nDEFAULT: stdout")

    args = parser.parse_args()

    start_time = datetime.datetime.now()
    #######################################
    #           logging setup             #
    #######################################
    logger = logging.getLogger("CAMSA.utils.agp2camsa_points")
    ch = logging.StreamHandler()
    ch.setLevel(args.c_logging_level)
    logger.setLevel(args.c_logging_level)
    logger.addHandler(ch)
    logger.info(full_description)
    logger.info(parser.format_values())
    ch.setFormatter(logging.Formatter(args.c_logging_formatter_entry))
    logger.info("Starting the converting process")

    objects = defaultdict(list)
    assembly_points = []

    if args.origin is None:
        logger.debug("\"origin\" were not specified explicitly for this AGP data. Inferring from the data stream")
        sources = []
        for source in args.agp:
            sources.append(os.path.splitext(os.path.basename(str(source.name)))[0])
        args.origin = ".".join(sources)
        logger.debug("\"origin\" has been inferred to \"{sources}\"".format(sources=args.origin))

    #######################################
    #      reading input AGP data         #
    #######################################
    logger.info("Reading input AGP formatted data")
    for source in args.agp:
        for line in source:
            line = line.strip()
            if line.startswith("#"):
                logger.debug("Skipping comment line: {line}".format(line=line))
                continue
            data = line.split("\t", 8)
            if Component.is_scaffold_component(data[4]):
                logger.debug("Processing a non-gap data line: {line}".format(line=line))
                component = ScaffoldComponent.from_agp_data(data=data)
            else:
                logger.debug("Processing a gap data line: {line}".format(line=line))
                component = GapComponent.from_agp_data(data=data)
            objects[component.object_id].append(component)

    for object_id in objects.keys():
        logger.debug("Processing object {object_id}".format(object_id=object_id))
        objects[object_id] = sorted(objects[object_id], key=lambda component: (component.object_beg, component.object_end))
        components = objects[object_id]
        camsa_points = object_as_camsa_points(object_as_components=components, extra_data=args)
        assembly_points.extend(camsa_points)

    logger.info("Writing CAMSA formatted assembly poitns to {file}".format(file=args.output.name))
    camsa_io.write_assembly_points(assembly_points=assembly_points, destination=args.output, output_setup=args.o_format)
    logger.info("Finished the conversion.")
    logger.info("Elapsed time: {el_time}".format(el_time=str(datetime.datetime.now() - start_time)))



from setuptools import setup
import sys
import os

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import camsa

setup(
    name='CAMSA',
    version=camsa.VERSION,
    author='Sergey Aganezov',
    author_email='aganezov@gwu.edu',
    description='CAMSA: a tool for Comparative Analysis and Merging of Scaffold Assemblies',
    license='MIT',
    keywords="comparative genomics, scaffolding, genome assembly",
    url='https://github.com/compbiol/camsa',
    packages=['', 'camsa', 'camsa.core', 'camsa.utils', 'camsa.utils.camsa', 'camsa.utils.fasta', 'camsa.utils.agp', 'camsa.utils.ragout', 'camsa.utils.grimm'],
    include_package_data=True,
    install_requires=['six>=1.10.0', 'networkx>=1.11', 'Jinja2>=2.8', 'enum34>=1.1.6', 'blist>=1.3.6', 'ConfigArgParse>=0.10.0',
                      'biopython>=1.67', 'bg>=1.8.1'],
    scripts=["camsa/run_camsa.py",
             "camsa/utils/ragout/ragout_coords2fasta.py", "camsa/utils/ragout/ragout_coords_coverage.py", "camsa/utils/ragout/ragout_coords2camsa_seqi.py", "camsa/utils/ragout/ragout_coords2camsa_points.py",
             "camsa/utils/grimm/grimm2camsa_points.py",
             "camsa/utils/fasta/fasta2camsa_points.py", "camsa/utils/fasta/fasta2camsa_seqi.py", "camsa/utils/fasta/camsa_points2fasta.py",
             "camsa/utils/agp/agp2camsa_points.py"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    long_description=open("pypi_full.rst").read()

)



# -*- coding: utf-8 -*-
import enum
import itertools

import networkx

from camsa.core.data_structures import ScaffoldAssemblyGraph


class Conflicts(enum.Enum):
    non_conflicted = 0
    semi_conflicted = 1
    conflicted = 2


def edges_conflict(edge1, edge2):
    u1, v1 = edge1
    u2, v2 = edge2
    return (u1 in [u2, v2] and v1 not in [u2, v2]) or (u1 not in [u2, v2] and v1 in [u2, v2])


def get_conflict_type(ap1, ap2):
    conflicted_cnt = 0
    non_conflicted_cnt = 0
    for edge1, edge2 in itertools.product(ap1.get_edges(), ap2.get_edges()):
        if edges_conflict(edge1=edge1, edge2=edge2):
            conflicted_cnt += 1
        else:
            non_conflicted_cnt += 1
    if conflicted_cnt == 0:
        return Conflicts.non_conflicted
    elif conflicted_cnt > 0 and non_conflicted_cnt > 0:
        return Conflicts.semi_conflicted
    else:
        return Conflicts.conflicted


def get_conflicting_edges(sag, ap_edge):
    result = []
    for vertex in ap_edge:
        conflicting_edges = [edge for edge in sag.graph.edges(nbunch=vertex, data=True)
                             if edges_conflict(edge1=(edge[0], edge[1]), edge2=ap_edge)]
        result.extend(conflicting_edges)
    return result


def get_conflicting_assembly_points(sag, assembly_point, assembly_points_by_ids):
    conflicting_assembly_points_ids = []
    for edge in assembly_point.get_edges():
        conflicting_edges = get_conflicting_edges(sag=sag, ap_edge=edge)
        ids = [data["ap_id"] for (u, v, data) in conflicting_edges]
        conflicting_assembly_points_ids.extend(ids)
    return [assembly_points_by_ids[ap_id] for ap_id in set(conflicting_assembly_points_ids) if ap_id != assembly_point.self_id]


def update_conflicts(ap1, ap2, in_ap1, out_ap1, in_ap2, out_ap2):
    in_sources = set(ap1.sources) & set(ap2.sources)
    ap1_out_sources = set(ap2.sources) - set(ap1.sources)
    ap2_out_sources = set(ap1.sources) - set(ap2.sources)
    for source in in_sources:
        in_ap1[source].add(ap2.self_id)
        in_ap2[source].add(ap1.self_id)
    for source in ap1_out_sources:
        out_ap1[source].add(ap2.self_id)
    for source in ap2_out_sources:
        out_ap2[source].add(ap1.self_id)


def update_assembly_points_as_conflicted(ap1, ap2):
    in_ap1_destination = ap1.in_conflicted
    out_ap1_destination = ap1.out_conflicted
    in_ap2_destination = ap2.in_conflicted
    out_ap2_destination = ap2.out_conflicted
    update_conflicts(ap1=ap1, ap2=ap2,
                     in_ap1=in_ap1_destination, out_ap1=out_ap1_destination,
                     in_ap2=in_ap2_destination, out_ap2=out_ap2_destination)


def update_assembly_points_as_semi_conflicted(ap1, ap2):
    in_ap1_destination = ap1.in_semi_conflicted
    out_ap1_destination = ap1.out_semi_conflicted
    in_ap2_destination = ap2.in_semi_conflicted
    out_ap2_destination = ap2.out_semi_conflicted
    update_conflicts(ap1=ap1, ap2=ap2,
                     in_ap1=in_ap1_destination, out_ap1=out_ap1_destination,
                     in_ap2=in_ap2_destination, out_ap2=out_ap2_destination)


def compute_and_update_assembly_points_conflicts(assembly_points_by_ids):
    """

    :param assembly_points_by_ids: a dictionary, where key is the assembly point id, and value is merged assembly points
    :return:
    """
    sag = ScaffoldAssemblyGraph.from_assembly_points(assembly_points=assembly_points_by_ids.values(),
                                                     add_scaffold_edges=False)
    for ap in assembly_points_by_ids.values():
        conflicted_assembly_points = get_conflicting_assembly_points(sag=sag, assembly_point=ap,
                                                                     assembly_points_by_ids=assembly_points_by_ids)
        for c_ap in conflicted_assembly_points:
            conflict_type = get_conflict_type(ap1=ap, ap2=c_ap)
            assert conflict_type != Conflicts.non_conflicted
            if conflict_type == Conflicts.conflicted:
                update_assembly_points_as_conflicted(ap1=ap, ap2=c_ap)
            elif conflict_type == Conflicts.semi_conflicted:
                update_assembly_points_as_semi_conflicted(ap1=ap, ap2=c_ap)



# -*- coding: utf-8 -*-


def get_time_string(start, end):
    hours, rem = divmod(end - start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)
